# Copyright 2020-2022 Laurent Cabaret
# Copyright 2020-2022 Vincent Jacques

from __future__ import annotations

from typing import List, Optional, Tuple
import csv
import dataclasses
import glob
import json
import os
import shlex
import unittest

import dacite


dataclass = dataclasses.dataclass(kw_only=True, frozen=True, slots=True)


# WARNING: if you make changes in these classes after version 1.0.0 is published,
# you may have to change the `format_version` in `RunResults.save`.

@dataclass
class MemoryInstantMetrics:
    rss: float


@dataclass
class InputOutputoInstantMetrics:
    read_chars: int
    write_chars: int


@dataclass
class ContextSwitchInstantMetrics:
    voluntary: int
    involuntary: int


@dataclass
class ProcessInstantMetrics:
    timestamp: float
    threads: int
    cpu_percent: float
    user_time: float
    system_time: float
    memory: MemoryInstantMetrics
    open_files: int
    io: InputOutputoInstantMetrics
    context_switches: ContextSwitchInstantMetrics
    gpu_percent: Optional[float]
    gpu_memory: Optional[float]


@dataclass
class ChroneEvent:
    process_id: str
    thread_id: str
    timestamp: int


@dataclass
class StopwatchStart(ChroneEvent):
    function_name: str
    label: Optional[str]
    index: Optional[int]


@dataclass
class StopwatchStop(ChroneEvent):
    pass


@dataclass
class StopwatchSummary(ChroneEvent):
    function_name: str
    label: Optional[str]
    executions_count: int
    average_duration: int
    duration_standard_deviation: int
    min_duration: int
    median_duration: int
    max_duration: int
    total_duration: int


@dataclass
class Process:
    command_list: List[str]
    pid: int
    started_between_timestamps: Tuple[float, float]
    terminated_between_timestamps: Tuple[float, float]
    instant_metrics: List[ProcessInstantMetrics]
    children: List[Process]

    @property
    def command(self):
        return shlex.join(self.command_list)

    def load_chrone_events(self):
        return load_chrone_events(self.pid)


@dataclass
class MainProcessGlobalMetrics:
    user_time: float
    system_time: float
    minor_page_faults: int
    major_page_faults: int
    input_blocks: int
    output_blocks: int
    voluntary_context_switches: int
    involuntary_context_switches: int


@dataclass
class MainProcess(Process):
    exit_code: int
    global_metrics: MainProcessGlobalMetrics


@dataclass
class SystemInstantMetrics:
    timestamp: float
    host_to_device_transfer_rate: Optional[float]
    device_to_host_transfer_rate: Optional[float]


@dataclass
class System:
    instant_metrics: List[SystemInstantMetrics]


@dataclass
class RunSettings:
    gpu_monitored: bool


@dataclass
class RunResults:
    run_settings: RunSettings
    system: System
    main_process: MainProcess

    def save(self, logs_dir):
        data = dataclasses.asdict(self)
        # If you have to change the format_version, you *must* add an integration test
        # with data in the previous format, to prove we can still load it.
        file_content = dict(format_version=1, data=data)
        with open(os.path.join(logs_dir, "run-result.json"), "w") as f:
            json.dump(file_content, f)

    @classmethod
    def load(cls) -> RunResults:
        with open("run-result.json") as f:
            file_content = json.load(f)
        assert file_content["format_version"] == 1
        data = file_content["data"]
        return dacite.from_dict(data_class=cls, data=data, config=dacite.Config(cast=[Tuple]))


def load_chrone_events(pid):
    chrones_file_names = glob.glob(f"*.{pid}.chrones.csv")
    if len(chrones_file_names) != 1:
        return

    with open(chrones_file_names[0]) as f:
        for line in csv.reader(f):
            yield make_chrone_event(line)


def make_chrone_event(line):
    process_id = line[0]
    thread_id = line[1]
    timestamp = int(line[2]) / 1e9
    if line[3] == "sw_start":
        return StopwatchStart(
            process_id=process_id,
            thread_id=thread_id,
            timestamp=timestamp,
            function_name=line[4],
            label=None if line[5] == "-" else line[5],
            index=None if line[6] == "-" else int(line[6]),
        )
    elif line[3] == "sw_stop":
        return StopwatchStop(
            process_id=process_id,
            thread_id=thread_id,
            timestamp=timestamp,
        )
    elif line[3] == "sw_summary":
        return StopwatchSummary(
            process_id=process_id,
            thread_id=thread_id,
            timestamp=timestamp,
            function_name=line[4],
            label=None if line[5] == "-" else line[5],
            executions_count=int(line[6]),
            average_duration=int(line[7]),
            duration_standard_deviation=int(line[8]),
            min_duration=int(line[9]),
            median_duration=int(line[10]),
            max_duration=int(line[11]),
            total_duration=int(line[12]),
        )
    else:
        assert False


class MakeChroneEventTestCase(unittest.TestCase):
    def test_stopwatch_start(self):
        self.assertEqual(
            make_chrone_event(["process_id", "thread_id", "375", "sw_start", "function_name", "label", "0"]),
            StopwatchStart(
                process_id="process_id",
                thread_id="thread_id",
                timestamp=375e-9,
                function_name="function_name",
                label="label",
                index=0,
            ),
        )

    def test_stopwatch_start_no_index(self):
        self.assertEqual(
            make_chrone_event(["process_id", "thread_id", "375", "sw_start", "function_name", "label", "-"]),
            StopwatchStart(
                process_id="process_id",
                thread_id="thread_id",
                timestamp=375e-9,
                function_name="function_name",
                label="label",
                index=None,
            ),
        )

    def test_stopwatch_start_no_label(self):
        self.assertEqual(
            make_chrone_event(["process_id", "thread_id", "375", "sw_start", "function_name", "-", "-"]),
            StopwatchStart(
                process_id="process_id",
                thread_id="thread_id",
                timestamp=375e-9,
                function_name="function_name",
                label=None,
                index=None,
            ),
        )

    def test_stopwatch_stop(self):
        self.assertEqual(
            make_chrone_event(["process_id", "thread_id", "375", "sw_stop"]),
            StopwatchStop(
                process_id="process_id",
                thread_id="thread_id",
                timestamp=375e-9,
            ),
        )

    def test_stopwatch_summary(self):
        self.assertEqual(
            make_chrone_event(["process_id", "thread_id", "375", "sw_summary", "function_name", "label", 10, 9, 8, 7, 6, 5, 4]),
            StopwatchSummary(
                process_id="process_id",
                thread_id="thread_id",
                timestamp=375e-9,
                function_name="function_name",
                label="label",
                executions_count=10,
                average_duration=9,
                duration_standard_deviation=8,
                min_duration=7,
                median_duration=6,
                max_duration=5,
                total_duration=4,
            )
        )

    def test_stopwatch_summary_no_label(self):
        self.assertEqual(
            make_chrone_event(["process_id", "thread_id", "375", "sw_summary", "function_name", "-", 10, 9, 8, 7, 6, 5, 4]),
            StopwatchSummary(
                process_id="process_id",
                thread_id="thread_id",
                timestamp=375e-9,
                function_name="function_name",
                label=None,
                executions_count=10,
                average_duration=9,
                duration_standard_deviation=8,
                min_duration=7,
                median_duration=6,
                max_duration=5,
                total_duration=4,
            )
        )




# Copyright 2020-2022 Laurent Cabaret
# Copyright 2020-2022 Vincent Jacques

from __future__ import annotations

from typing import Dict, Iterable, List, Optional, Tuple
import dataclasses

import matplotlib.pyplot as plt

from ..monitoring import result as monitoring_result


def make_graph(output_file):
    results = monitoring_result.RunResults.load()

    origin_timestamp = results.main_process.started_between_timestamps[0]

    gantt_grapher = GantGrapher(results)

    n = 10 if results.run_settings.gpu_monitored else 7
    fig, axes = plt.subplots(
        n, 1, squeeze=False,
        sharex=True,
        figsize=(12, 4 * n + gantt_grapher.get_height() / 10), layout="constrained",
        height_ratios=[gantt_grapher.get_height() / 15] + [1 for _ in range(n - 1)],
    )
    if results.run_settings.gpu_monitored:
        (
            (chrones_ax,),
            (cpu_ax,), (threads_ax,),
            (rss_ax,),
            (gpu_ax,), (gpu_mem_ax,), (gpu_transfers_ax,),
            (inputs_ax,), (outputs_ax,), (open_files_ax,)
        ) = axes
    else:
        (
            (chrones_ax,),
            (cpu_ax,), (threads_ax,),
            (rss_ax,),
            (inputs_ax,), (outputs_ax,), (open_files_ax,)
        ) = axes

    gantt_grapher.draw(chrones_ax)

    def plot_cpu(process: monitoring_result.Process):
        metrics = process.instant_metrics
        cpu_ax.plot(
            [m.timestamp - origin_timestamp for m in metrics],
            [m.cpu_percent for m in metrics],
            ".-",
            label=process.command[-30:],
        )

    iter_processes(results.main_process, before=plot_cpu)
    cpu_ax.legend()
    cpu_ax.set_ylim(bottom=0)
    cpu_ax.set_ylabel("CPU (%)")

    def plot_threads(process: monitoring_result.Process):
        metrics = process.instant_metrics
        threads_ax.plot(
            [m.timestamp - origin_timestamp for m in metrics],
            [m.threads for m in metrics],
            ".-",
            label=process.command[-30:],
        )

    iter_processes(results.main_process, before=plot_threads)
    threads_ax.legend()
    threads_ax.set_ylim(bottom=0)
    threads_ax.set_ylabel("Threads")

    def plot_rss(process: monitoring_result.Process):
        metrics = process.instant_metrics
        rss_ax.plot(
            [m.timestamp - origin_timestamp for m in metrics],
            [m.memory.rss / 1024. / 1024. for m in metrics],
            ".-",
            label=process.command[-30:],
        )

    iter_processes(results.main_process, before=plot_rss)
    rss_ax.legend()
    rss_ax.set_ylim(bottom=0)
    rss_ax.set_ylabel("Memory - RSS (MiB)")

    if results.run_settings.gpu_monitored:
        def plot_gpu(process: monitoring_result.Process):
            metrics = process.instant_metrics
            gpu_ax.plot(
                [m.timestamp - origin_timestamp for m in metrics],
                [m.gpu_percent for m in metrics],
                ".-",
                label=process.command[-30:],
            )

        iter_processes(results.main_process, before=plot_gpu)
        gpu_ax.legend()
        gpu_ax.set_ylim(bottom=0)
        gpu_ax.set_ylabel("GPU (%)")

        def plot_gpu_mem(process: monitoring_result.Process):
            metrics = process.instant_metrics
            gpu_mem_ax.plot(
                [m.timestamp - origin_timestamp for m in metrics],
                [m.gpu_memory for m in metrics],
                ".-",
                label=process.command[-30:],
            )

        iter_processes(results.main_process, before=plot_gpu_mem)
        gpu_mem_ax.legend()
        gpu_mem_ax.set_ylim(bottom=0)
        gpu_mem_ax.set_ylabel("GPU memory (MiB)")

        gpu_transfers_ax.plot([m.timestamp - origin_timestamp for m in results.system.instant_metrics], [m.host_to_device_transfer_rate for m in results.system.instant_metrics], 'o-', label="Host to device",)
        gpu_transfers_ax.plot([m.timestamp - origin_timestamp for m in results.system.instant_metrics], [m.device_to_host_transfer_rate for m in results.system.instant_metrics], 'o-', label="Device to host",)
        gpu_transfers_ax.set_ylabel("System-wide\nGPU transfers (MB/s)")
        gpu_transfers_ax.legend()

    def plot_inputs(process: monitoring_result.Process):
        metrics = process.instant_metrics
        inputs_ax.plot(
            [m.timestamp - origin_timestamp for m in metrics],
            [m.io.read_chars / 1024. / 1024. for m in metrics],
            ".-",
            label=process.command[-30:],
        )

    iter_processes(results.main_process, before=plot_inputs)
    inputs_ax.legend()
    inputs_ax.set_ylim(bottom=0)
    inputs_ax.set_ylabel("Cumulated inputs (MiB)")

    def plot_outputs(process: monitoring_result.Process):
        metrics = process.instant_metrics
        outputs_ax.plot(
            [m.timestamp - origin_timestamp for m in metrics],
            [m.io.write_chars / 1024. / 1024. for m in metrics],
            ".-",
            label=process.command[-30:],
        )

    iter_processes(results.main_process, before=plot_outputs)
    outputs_ax.legend()
    outputs_ax.set_ylim(bottom=0)
    outputs_ax.set_ylabel("Cumulated outputs (MiB)")

    def plot_open_files(process: monitoring_result.Process):
        metrics = process.instant_metrics
        open_files_ax.plot(
            [m.timestamp - origin_timestamp for m in metrics],
            [m.open_files for m in metrics],
            ".-",
            label=process.command[-30:],
        )

    iter_processes(results.main_process, before=plot_open_files)
    open_files_ax.legend()
    open_files_ax.set_ylim(bottom=0)
    open_files_ax.set_ylabel("Open files")

    axes[-1][0].set_xlim(left=0)
    axes[-1][0].set_xlabel("Time (s)")

    fig.savefig(output_file, dpi=120)
    plt.close(fig)


class GantGrapher:
    @dataclasses.dataclass
    class Thread:
        stack: List[monitoring_result.ChroneEvent]
        chrones: Dict[str, Tuple[float, float]]
        first_event: Optional[monitoring_result.ChroneEvent]
        last_event: Optional[monitoring_result.ChroneEvent]

    def __init__(self, results: monitoring_result.RunResults):
        self.__results = results
        self.__origin_timestamp = results.main_process.started_between_timestamps[0]
        self.__prepare()

    def __prepare(self):
        self.__processes = []
        iter_processes(self.__results.main_process, before=lambda p: self.__processes.append(p))
        self.__threads = {process.pid: self.__prepare_threads(process) for process in self.__processes}

    def __prepare_threads(self, process: monitoring_result.Process):
        threads = {}
        for event in process.load_chrone_events():
            thread = threads.setdefault(event.thread_id, GantGrapher.Thread([], {}, None, None))
            if thread.first_event is None:
                thread.first_event = event
            thread.last_event = event

            if event.__class__ == monitoring_result.StopwatchStart:
                thread.stack.append(event)
            elif event.__class__ == monitoring_result.StopwatchStop:
                start_event = thread.stack.pop()
                name = " - ".join(
                    str(part)
                    for part in filter(
                        lambda p: p is not None,
                        [start_event.function_name, start_event.label],
                    )
                )
                chrones = thread.chrones.setdefault(name, [])
                chrones.append((start_event, event))
            elif event.__class__ == monitoring_result.StopwatchSummary:
                pass
            else:
                assert False
        threads = list(threads.values())
        assert all(t.stack == [] for t in threads)

        return threads

    def get_height(self):
        return sum(
            2 + sum(3 + len(thread.chrones) for thread in self.__threads[process.pid])
            for process in self.__processes
        ) - 1

    def draw(self, ax):
        top_y = 0

        for process in self.__processes:
            start_x = (process.started_between_timestamps[0] + process.started_between_timestamps[1]) / 2 - self.__origin_timestamp
            end_x = (process.terminated_between_timestamps[0] + process.terminated_between_timestamps[1]) / 2 - self.__origin_timestamp
            width = end_x - start_x

            threads_height = sum(3 + len(thread.chrones) for thread in self.__threads[process.pid])
            process_height = 1 + threads_height

            ax.broken_barh([(start_x, width)], (top_y - process_height, process_height), color="#ff8f8f")
            ax.text(x=start_x, y=top_y - 0.5, s=process.command, ha="left", va="center")

            self.__plot_threads(top_y - 1, self.__threads[process.pid], ax)

            top_y -= 1 + process_height

        assert top_y + 1 == -self.get_height()

        ax.set_ylim(top=0, bottom=top_y + 1)
        ax.set_yticks([])

    def __plot_threads(self, top_y, threads: List[GantGrapher.Thread], ax):
        for (thread_index, thread) in enumerate(filter(lambda t: t.chrones, threads)):
            start_x = thread.first_event.timestamp - self.__origin_timestamp
            end_x = thread.last_event.timestamp - self.__origin_timestamp
            width = end_x - start_x

            thread_height = 2 + len(thread.chrones)

            ax.broken_barh([(start_x, width)], (top_y - thread_height, thread_height), color="#8fff8f")
            ax.text(x=start_x, y=top_y - 0.5, s=f"Thread {thread_index}", ha="left", va="center")

            self.__plot_chrones(start_x, top_y - 1, thread.chrones, ax)

            top_y -= 1 + thread_height

    def __plot_chrones(self, left_x, top_y, chrones, ax: plt.Axes):
        for (name, events) in sorted(chrones.items(), key=lambda kv: kv[1][0][0].timestamp):
            bars = [(start_event.timestamp - self.__origin_timestamp, stop_event.timestamp - start_event.timestamp) for (start_event, stop_event) in events]
            ax.broken_barh(bars, (top_y - 1, 1), color="#8f8fff", edgecolor="black")
            ax.text(x=left_x, y=top_y - 0.5, s=name, ha="left", va="center")

            top_y -= 1


def iter_processes(process, *, before=lambda _: None, after=lambda _: None):
    before(process)
    for child in process.children:
        iter_processes(child, before=before, after=after)
    after(process)



# Copyright 2020-2022 Laurent Cabaret
# Copyright 2020-2022 Vincent Jacques

from __future__ import annotations

import os
import sys
import textwrap

import click

# @todo(later) How could I refer to these as `instrumentation.shell` and `instrumentation.cpp`?
from .instrumentation import shell as shell_instrumentation
from .instrumentation import cpp as cpp_instrumentation
from .monitoring.runner import Runner
from .reporting.graph import make_graph


@click.group(help="Chrones is a software development tool to visualize runtime statistics about your program and correlate them with the phases of your program. Please visit https://github.com/jacquev6/Chrones for more details.")
def main():
    pass


@main.group(help="Everything related to instrumentation.")
def instrument():
    pass


@instrument.group(name="c++", help="Instrumentation of C++ programs.")
def cpp():
    pass


@cpp.command(help="Display the location of the C++ header(s). Use as 'g++ -I`chrones instrument c++ header-location`'.")
def header_location():
    print(cpp_instrumentation.location())


@instrument.group(help="Instrumentation of shell scripts.")
def shell():
    pass


@shell.command(help="Display the 'chrones_*' shell fonctions. Use as 'source <(chrones instrument shell enable PROGRAM_NAME)'.")
@click.argument("program-name")
def enable(*, program_name):
    for line in shell_instrumentation.enable(program_name):
        print(line)


@main.command(help=textwrap.dedent("""\
    Run a program under Chrones' monitoring.
    Add a '--' before the command if you need to pass options to the command.

    You *may* want to clear I/O caches just before running your command.
    See for example https://stackoverflow.com/a/25336215/905845.
"""))
@click.option("--logs-dir", default=".", help="Directory where instrumentation and monitoring logs will be stored.")
@click.option("--monitor-gpu", is_flag=True, help="Measure GPU usage.")
@click.option("--monitoring-interval", type=float, default=0.2, help="Interval between two consecutives measures, in seconds.")
@click.option("--allowed-missing-samples", type=int, default=1, help="Number of successive samples that can be skipped before Chrones start displaying warnings about \"slow monitoring\".")
@click.argument("command", nargs=-1, type=click.UNPROCESSED)
def run(
    *,
    logs_dir,
    monitor_gpu,
    monitoring_interval,
    allowed_missing_samples,
    command,
):
    runner = Runner(
        monitor_gpu=monitor_gpu,
        logs_directory=logs_dir,
        monitoring_interval=monitoring_interval,
        allowed_missing_samples=allowed_missing_samples,
    )
    result = runner.run(list(command))
    result.save(logs_dir)
    if result.main_process.exit_code != 0:
        sys.exit(result.main_process.exit_code)


@main.command(help="Create a human-readable image from monitoring logs.")
@click.option("--logs-dir", default=".", help="Directory containing instrumentation and monitoring logs.")
@click.option("--output-name", default="report.png", help="Output name for the report.")
def report(*, logs_dir, output_name):
    output_name = os.path.abspath(output_name)
    os.chdir(logs_dir)
    make_graph(output_name)
    # @todo(later) Restore summaries (cf. Original chrones-report.py file)



# Copyright 2020-2022 Laurent Cabaret
# Copyright 2020-2022 Vincent Jacques

from __future__ import annotations

import os.path


def location():
    # @todo(later) Switch to https://setuptools.pypa.io/en/latest/userguide/datafiles.html#accessing-data-files-at-runtime
    return os.path.dirname(__file__)



# Copyright 2020-2022 Laurent Cabaret
# Copyright 2020-2022 Vincent Jacques

from __future__ import annotations

# https://packaging.python.org/en/latest/

import setuptools


version = "1.0.0"

with open("README.md") as f:
    long_description = f.read()
long_description = long_description.replace("integration-tests/readme-example/report.png", f"https://github.com/jacquev6/Chrones/raw/v{version}/integration-tests/readme-example/report.png")

with open("requirements.txt") as f:
    install_requires = f.readlines()

setuptools.setup(
    name="Chrones",
    version=version,
    description="Software development tool to visualize runtime statistics about your program and correlate them with its phases",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jacquev6/Chrones",
    author="Vincent Jacques",
    author_email="vincent@vincent-jacques.net",
    license="MIT",
    install_requires=install_requires,
    packages=setuptools.find_packages(),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "chrones = Chrones.command_line_interface:main",
        ],
    },
)
